@@ -1,7 +1,7 @@
 # Automated tests for the `executor' module.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: March 2, 2020
+# Last Change: May 14, 2020
 # URL: https://executor.readthedocs.io
 
 """
@@ -56,10 +56,11 @@ import uuid
 
 # External dependencies.
 from humanfriendly import Timer, coerce_boolean
-from humanfriendly.testing import TemporaryDirectory, TestCase, retry, run_cli
+from humanfriendly.testing import MockedProgram, TemporaryDirectory, TestCase, retry, run_cli
 from humanfriendly.text import compact, dedent
 from mock import MagicMock
 from property_manager import set_property
+from six import text_type
 from six.moves import StringIO
 
 # Modules included in our package.
@@ -99,6 +100,7 @@ from executor.ssh.client import (
 from executor.ssh.server import SSHServer
 
 MISSING_COMMAND = 'a-program-name-that-no-one-would-ever-use'
+EXPECTED_CODENAMES = (u'xenial', u'bionic', u'focal')
 
 # Initialize a logger for this module.
 logger = logging.getLogger(__name__)
@@ -1035,55 +1037,76 @@ class ExecutorTestCase(TestCase):
         assert create_context(sudo=True).options['sudo'] is True
         assert create_context(sudo=False).options['sudo'] is False
 
+    def test_apt_sources_info(self):
+        """Test parsing of ``/etc/apt/sources.list`` for release discovery."""
+        with SwallowReleaseDiscoveryErrors():
+            context = LocalContext()
+            # Mock read_file() to feed expected data to 'apt_sources_info'.
+            sources_entry = u'deb http://archive.ubuntu.com/ubuntu/ focal main restricted'
+            context.read_file = MagicMock(return_value=sources_entry.encode('UTF-8'))
+            # Verify that 'apt_sources_info' successfully determines the
+            # distributor ID and distribution codename on Ubuntu systems.
+            assert context.apt_sources_info[0] == u'ubuntu'
+            assert context.apt_sources_info[1] == u'focal'
+
+    def test_apt_sources_info_fallback(self):
+        """Test that release discovery falls back to ``/etc/apt/sources.list`` parsing."""
+        with SwallowReleaseDiscoveryErrors():
+            # Shadow the system wide 'lsb_release' program with a shell script
+            # that simply exits with status 255, to simulate the 'lsb_release'
+            # program not being available.
+            with MockedProgram(name='lsb_release', returncode=255):
+                context = LocalContext()
+                # Disable parsing of the /etc/lsb-release file.
+                set_property(context, 'lsb_release_variables', {})
+                # Mock read_file() to feed expected data to 'apt_sources_info'.
+                sources_entry = u'deb http://deb.debian.org/debian/ buster main'
+                context.read_file = MagicMock(return_value=sources_entry.encode('UTF-8'))
+                # Verify that 'distributor_id' and 'distribution_codename' fall
+                # back to the information gathered by 'apt_sources_info'.
+                assert context.distributor_id == u'debian'
+                assert context.distribution_codename == u'buster'
+
     def test_lsb_release_shortcuts(self):
         """Test the ``lsb_release`` shortcuts."""
-        # The following tests should pass on my laptops and Travis CI, but I
-        # don't want this test to fail on `unexpected' platforms, so here's a
-        # pragmatic compromise :-).
-        try:
+        with SwallowReleaseDiscoveryErrors():
             context = LocalContext()
-            assert context.distributor_id == 'ubuntu'
-            assert context.distribution_codename in ('precise', 'trusty', 'xenial', 'bionic')
-        except Exception:
-            if self.lsb_release_expected_to_work:
-                raise
-            else:
-                return self.skipTest("assuming unsupported platform")
+            assert context.distributor_id == u'ubuntu'
+            assert context.distribution_codename in EXPECTED_CODENAMES
 
     def test_lsb_release_fallback(self):
         """Make sure the fall back from /etc/lsb-release to /usr/bin/lsb_release works."""
-        context = LocalContext()
-        # Disable the parsing of /etc/lsb-release.
-        set_property(context, 'lsb_release_variables', {})
-        # Test the fall back from /etc/lsb-release to /usr/bin/lsb_release.
-        try:
-            assert context.distributor_id == 'ubuntu'
-            assert context.distribution_codename in ('precise', 'trusty', 'xenial', 'bionic')
-        except Exception:
-            if self.lsb_release_expected_to_work:
-                raise
-            else:
-                return self.skipTest("assuming unsupported platform")
+        with SwallowReleaseDiscoveryErrors():
+            context = LocalContext()
+            # Disable parsing of the /etc/lsb-release file.
+            set_property(context, 'lsb_release_variables', {})
+            # Test the fall back from /etc/lsb-release to /usr/bin/lsb_release.
+            assert context.distributor_id == u'ubuntu'
+            assert context.distribution_codename in EXPECTED_CODENAMES
 
     def test_lsb_release_error_handling(self):
         """Test that the ``lsb_release`` shortcuts don't raise exceptions on unsupported platforms."""
         context = LocalContext(environment=dict(PATH=''))
-        assert context.distributor_id == ''
-        assert context.distribution_codename == ''
+        assert context.distributor_id == u''
+        assert context.distribution_codename == u''
 
     def test_lsb_release_variables(self):
         """Test the ``/etc/lsb-release`` parsing."""
-        try:
+        with SwallowReleaseDiscoveryErrors():
             context = LocalContext()
             variables = context.lsb_release_variables
             assert len(variables) > 0
-            assert variables['DISTRIB_ID'].lower() == 'ubuntu'
-            assert variables['DISTRIB_CODENAME'].lower() in ('precise', 'trusty', 'xenial', 'bionic')
-        except Exception:
-            if self.lsb_release_expected_to_work:
-                raise
-            else:
-                return self.skipTest("assuming unsupported platform")
+            # NB: The following tests are written to assume one of the
+            # supported Ubuntu LTS releases, because the developer's
+            # laptop as well as Travis CI run Ubuntu LTS releases.
+            assert variables['DISTRIB_ID'].lower() == u'ubuntu'
+            assert variables['DISTRIB_CODENAME'].lower() in EXPECTED_CODENAMES
+            # Make sure all strings are Unicode strings, this tests against
+            # a regression of a bug caused by shlex.split() on Python 2.7
+            # automatically coercing Unicode strings to byte strings.
+            for key, value in context.lsb_release_variables.items():
+                assert isinstance(key, text_type)
+                assert isinstance(value, text_type)
 
     def test_local_context(self):
         """Test a local command context."""
@@ -1255,17 +1278,6 @@ class ExecutorTestCase(TestCase):
             assert timer.elapsed_time < 10
         retry(timeout_hammer, 60)
 
-    @property
-    def lsb_release_expected_to_work(self):
-        """True on my laptops and Travis CI, False otherwise."""
-        developer_laptop = (socket.gethostname() in ('peter-mba', 'peter-mbp'))
-        travis_ci = coerce_boolean(os.environ.get('TRAVIS', 'false'))
-        # TODO For some weird reason the parsing of the /etc/lsb-release file
-        # in the Travis CI build environment for Python 2.6 seems to result in
-        # binary garbage. Right now I'm not inclined to dive deeper.
-        py26 = (sys.version_info[:2] == (2, 6))
-        return developer_laptop or (travis_ci and not py26)
-
 
 def intercept(exc_type, func, *args, **kw):
     """Intercept and return a raised exception."""
@@ -1311,3 +1323,23 @@ class NonForcefulCommand(NonGracefulCommand):
             disabled to simulate processes that refuse to terminate
             forcefully ..
         """))
+
+
+class SwallowReleaseDiscoveryErrors(object):
+
+    """Context manager to conditionally swallow release discovery errors."""
+
+    def __enter__(self):
+        """Enable use as a context manager."""
+        return self
+
+    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
+        """Swallow release discovery errors (except on the developer's laptop and Travis CI)."""
+        if exc_type is AssertionError:
+            developer_laptop = (socket.gethostname() in ('peter-mba', 'peter-mbp'))
+            travis_ci = coerce_boolean(os.environ.get('TRAVIS', 'false'))
+            if developer_laptop or travis_ci:
+                # Re-raise the assertion failure.
+                return False
+            else:
+                self.skipTest("assuming unsupported platform")
